/*
 *   ScenarLang a domain specific language that attempts to make it even easier
 *   to write unit tests with mockito
 *    
 *   Copyright (C) 2012  X. Gillard 
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *   
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package xgi.ut.dsl.utl;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.mockito.cglib.proxy.Enhancer;

import xgi.ut.dsl.annot.ConversionMethod;
import xgi.ut.dsl.utl.filtering.AndFilter;
import xgi.ut.dsl.utl.filtering.Filter;
import xgi.ut.dsl.utl.filtering.IsStaticFilter;
import xgi.ut.dsl.utl.filtering.SignatureMatchFilter;
import xgi.ut.dsl.utl.matching.MetaMatcher;
import xgi.ut.dsl.utl.mock.MockInvocation;
import xgi.ut.dsl.utl.mock.SetMockResult;
import xgi.ut.dsl.utl.mock.ThrowException;
import xgi.ut.dsl.utl.overloading.ConstructorSpecificityComparator;
import xgi.ut.dsl.utl.overloading.MethodSpecificityComparator;
/**
 * This class provides a convenient workaround to avoid the fact that 
 * scala arrays are not automatically expanded to vararg when it comes
 * to java.lang.reflect.Method invocation
 */
public class ReflectionUtil {
	/** This field holds a cache of all the fields of all the previously inspected classes */
	private static final Map<Class<?>, Map<String, Field>> FIELD_CACHE = new HashMap<Class<?>, Map<String,Field>>();
	/** This field holds a cache of all the methods of all the previously inspected classes */
	private static final Map<Class<?>, Map<String, List<Method>>> METHOD_CACHE = new HashMap<Class<?>, Map<String,List<Method>>>();
	/**	Pertains the assignable from relationship */
	private static final Map<Class, List<Class>> BOXING_MATRIX = new HashMap<Class, List<Class>>();
	static{
		BOXING_MATRIX.put(double.class, Arrays.asList(new Class[]{char.class, byte.class, short.class, int.class, long.class, float.class, double.class}));
		BOXING_MATRIX.put(float.class, Arrays.asList( new Class[]{char.class, byte.class, short.class, int.class, long.class, float.class}));
		BOXING_MATRIX.put(long.class, Arrays.asList(  new Class[]{char.class, byte.class, short.class, int.class, long.class}));
		BOXING_MATRIX.put(int.class, Arrays.asList(   new Class[]{char.class, byte.class, short.class, int.class}));
		BOXING_MATRIX.put(short.class, Arrays.asList( new Class[]{char.class, byte.class, short.class}));
		BOXING_MATRIX.put(byte.class, Arrays.asList(  new Class[]{char.class, byte.class}));
		BOXING_MATRIX.put(boolean.class, Arrays.asList(new Class[]{boolean.class}));
	}
	
	/**
	 * This will produce an instance of class using a constructor that accepts the given args
	 * @param clazz the class to instanciate
	 * @param args the constructor arguments 
	 * @return an instance of clazz
	 * @throws IllegalArgumentException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	public static final Object instanciate(Class<?> clazz, List<?> args) throws IllegalArgumentException, InstantiationException, IllegalAccessException, InvocationTargetException {
		Constructor<?> constr = getMostSpecificConstructor(clazz, args);
		constr.setAccessible(true);
		return constr.newInstance(boxArgs(args.toArray(), constr.getParameterTypes()));
	}
	/**
	 * Retrieves a constructor that has a signature compatible with the given parameters
	 * @param clazz the class to inspect
	 * @param name the method name
	 * @param args the method arguments
	 * @return the method
	 */
	public static Constructor<?> getMostSpecificConstructor(Class<?>clazz, List<?> args) {
		Constructor<?>[] all = clazz.getDeclaredConstructors();
		if(all == null || all.length == 0){
			throw new RuntimeException("No such constructor "+clazz.getCanonicalName()+"("+Arrays.asList(args.toArray())+")");
		}
		
		List<Constructor<?>> matching = new ArrayList<Constructor<?>>();
		for(Constructor<?> constr : all) {
			Class<?>[] paramTypes = constr.getParameterTypes();
			if(argumentsMatch(paramTypes, args)){
				matching.add(constr);
			}
		}
		
		if(matching.isEmpty()){
			throw new RuntimeException("No such constructor "+clazz.getCanonicalName()+"("+Arrays.asList(args.toArray())+")");
		}
		
		ConstructorSpecificityComparator cmp = new ConstructorSpecificityComparator();
		Collections.sort(matching, cmp);
		
		if(matching.size()>1){
			Constructor<?> _1st = matching.get(0);
			Constructor<?> _2nd = matching.get(1);
			if(cmp.compare(_1st, _2nd)==0){
				throw new RuntimeException(String.format(
								"I cannot choose the proper constructor to execute : it's ambiguous between at least \n%1$s\n%2$s ", 
								_1st, _2nd));
			}
		}
		return matching.get(0);
	}
	
	/** 
	 * invokes the method identified by name on object target with the given parameters 
	 * and assigns it a mock result.
	 * @param target the object on which to invoke the method
	 * @param name the name of the method to invoke
	 * @param params the method parameters
	 * @param result the mock result to assign to the fn call
	 * @return the result of the fn call
	 * @throws NoSuchMethodException in case the method does not exists in the class
	 * @throws IllegalAccessException in case the method is not accessible
	 * @throws InvocationTargetException in case the method cannot be called
	 */
	public static final Object invoke(Object target, String name,
			List<?> params) throws NoSuchMethodException,
			IllegalAccessException, InvocationTargetException {
		Class<?> clazz = target.getClass();
		// retrieve the method
		Method m = getMostSpecificMatch(clazz, name, params, null);

		return m.invoke(target, boxArgs(params.toArray(), m.getParameterTypes()));
	}
	
	/** 
	 * invokes the method identified by name on object target with the given parameters 
	 * and assigns it a mock result.
	 * @param target the object on which to invoke the method
	 * @param name the name of the method to invoke
	 * @param params the method parameters
	 * @param result the mock result to assign to the fn call
	 * @return the result of the fn call
	 * @throws NoSuchMethodException in case the method does not exists in the class
	 * @throws IllegalAccessException in case the method is not accessible
	 * @throws InvocationTargetException in case the method cannot be called
	 */
	public static final Object invokeStatic(Class<?> clazz, String name, List<?> params) throws NoSuchMethodException, IllegalAccessException, InvocationTargetException {
		// retrieve the method
		List<Method> possibles = getAllMethods(clazz).get(name);
		if(possibles==null || possibles.isEmpty()){
			throw new RuntimeException(String.format("No static method is found in class %s with name %s (at all)", clazz.getCanonicalName(), name));
		}
		
		Method m = getMostSpecificStaticMatch(clazz, name, params, null);
		return m.invoke(null, boxArgs(params.toArray(), m.getParameterTypes()));
	}
	
	/** 
	 * invokes the method identified by name on object target with the given parameters 
	 * and assigns it a mock result.
	 * @param target the object on which to invoke the method
	 * @param name the name of the method to invoke
	 * @param params the method parameters
	 * @param result the mock result to assign to the fn call
	 * @return the result of the fn call
	 * @throws NoSuchMethodException in case the method does not exists in the class
	 * @throws IllegalAccessException in case the method is not accessible
	 * @throws InvocationTargetException in case the method cannot be called
	 */
	public static final Object invokeResult(
			Object target, 
			String name,
			List<? extends MetaMatcher> params, 
			Object result) throws NoSuchMethodException, IllegalAccessException, InvocationTargetException {
		
		return invokeOnMock(target, name, params, new SetMockResult(result), "You cannot configure a result on an object that is not a mock");
	}
	
	/** 
	 * invokes the method identified by name on object target with the given parameters 
	 * and assigns it a mock result.
	 * @param target the object on which to invoke the method
	 * @param name the name of the method to invoke
	 * @param params the method parameters
	 * @param result the mock result to assign to the fn call
	 * @return the result of the fn call
	 * @throws NoSuchMethodException in case the method does not exists in the class
	 * @throws IllegalAccessException in case the method is not accessible
	 * @throws InvocationTargetException in case the method cannot be called
	 */
	public static final void invokeException(
			Object target, 
			String name,
			List<? extends MetaMatcher> params, 
			Throwable result) throws NoSuchMethodException,	IllegalAccessException, InvocationTargetException {
		
		invokeOnMock(target, name, params, new ThrowException(result), "You cannot configure an exception on an object that is not a mock");
	}
	
	private static final Object invokeOnMock(
			Object target, 
			String name, 
			List<? extends MetaMatcher> params, 
			MockInvocation invocation, 
			String errorText){
		
		Class<?> clazz = target.getClass();
		
		List<Class<?>> actualArgsType = new ArrayList<Class<?>>();
		for(MetaMatcher matcher : params){
			actualArgsType.add(matcher.getDeclaredType());
		}
		
		Method m = getMostSpecificMatchFromFilter(clazz, name, new SignatureMatchFilter(name, actualArgsType, null));
		
		Object retVal = null;
		// Invoke the method		
		if (!Enhancer.isEnhanced(clazz)) {// was original class a stub ? 
			throw new IllegalStateException(errorText);
		} else {
			try {
				retVal = invocation.invoke(target, m, params);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		return retVal;
	}
	
	/**
	 * This method provides a convenient way to invoke the conversion method of a conversion class
	 * @param text the text that needs to be converted (parsed)
	 * @param conversionClass the class that contains one conversion method (public static annotated with @ConversionMethod)
	 * @return the object that is represented by text
	 * 
	 * @throws IllegalArgumentException in case there is a mismatch btw the given arguments and that of the conversion method
	 */
	public static final Object invokeConversion(String text, Class<?> conversionClass) throws IllegalArgumentException {
		Method[] methods = conversionClass.getMethods();
		for(Method m : methods){
			ConversionMethod annotation = m.getAnnotation(ConversionMethod.class);
			if(annotation!=null){
				if(!Modifier.isStatic(m.getModifiers()))throw new RuntimeException("The conversion method must be static");
				if(!Modifier.isPublic(m.getModifiers()))throw new RuntimeException("The conversion method must be public");
				try {
					return m.invoke(null, text);
				} catch (IllegalAccessException e) {
					//swallow : it cannot happen
				} catch (InvocationTargetException e) {
					//swallow : it cannot happen
				}
			}
		}
		throw new RuntimeException("The parser class "+conversionClass.getCanonicalName()+" contains no public static method annotated with @ConversionMethod");
	}
	
	/**
	 * Retrieves all the public + private (declared + inherited) fields of a given class
	 * 
	 * Note : as a side effect, this method will populate the CACHE field of this class.
	 * Note2: as a side effect, this will make the field accessible.
	 * 
	 * @param clazz the class to inspect
	 * @return a map with the key being the field name and the value the field itself.
	 */
	public static Map<String, Field> getAllFields(Class<?> clazz){
		Map<String, Field> results = FIELD_CACHE.get(clazz);
		if(results!=null){
			return results;
		}
		if(clazz.equals(Object.class)){
			return new HashMap<String, Field>();
		}
		Class<?> parent = clazz.getSuperclass();
		results = new HashMap<String, Field>(getAllFields(parent));
		for(Field f : clazz.getDeclaredFields()){
			f.setAccessible(true);
			results.put(f.getName(), f);
		}
		FIELD_CACHE.put(clazz, results);
		return results;
	}
	
	/**
	 * Retrieves all the public + private (declared + inherited) methods of a given class
	 * 
	 * Note : as a side effect, this method will populate the method CACHE of this class.
	 * Note2: mock classes are not taken into account
	 * Note3: as a side effect, this method will make the methods accessible
	 * 
	 * @param clazz the class to inspect
	 * @return a map with the key being the method name and the value the list of all methods named the same way.
	 */
	public static Map<String, List<Method>> getAllMethods(Class<?> clazz){
		Map<String, List<Method>> results = METHOD_CACHE.get(clazz);
		if(results!=null){
			return results;
		}
		
		Class<?> parent = clazz.getSuperclass();
		results = new HashMap<String, List<Method>>();
		if(parent!=null){
			results.putAll(getAllMethods(parent));
		}
		for(Method m : clazz.getDeclaredMethods()){
			List<Method> methods = results.get(m.getName());
			if(methods == null){
				methods = new ArrayList<Method>();
				results.put(m.getName(), methods);
			}
			m.setAccessible(true);
			methods.add(m);
		}
		METHOD_CACHE.put(clazz, results);
		return results;
	}
	
	public static Method getMostSpecificMatch(Class<?> clazz, String name, List<?> args, Object result){
		Filter<Method> filter = new SignatureMatchFilter(name, classesOf(args), result==null ? null : result.getClass());
		return getMostSpecificMatchFromFilter(clazz, name, filter);
	}
	
	public static Method getMostSpecificStaticMatch(Class<?> clazz, String name, List<?> args, Object result){
		Filter<Method> matches = new SignatureMatchFilter(name, classesOf(args), result==null ? null : result.getClass());
		Filter<Method> isStatic = new IsStaticFilter();
		return getMostSpecificMatchFromFilter(clazz, name, new AndFilter<Method>(matches, isStatic));
	}
	
	private static Method getMostSpecificMatchFromFilter(Class<?> clazz, String name, Filter<Method> matchFilter){
		List<Method> applicables = filterMethods(clazz, name, matchFilter);
		Comparator<Method> cmp = new MethodSpecificityComparator();
		Collections.sort(applicables, cmp);
		
		if(applicables.size()==0){
			String message = "%s not found in class %s\n" +
			"You might want to double check : \n" +
			"* spelling mistakes \n" +
			"* wrong arguments type\n" +
			"* wrong return type\n";
			throw new RuntimeException(String.format(message, matchFilter.toString(), clazz));
		}
		
		if(applicables.size()>1){
			Method _1st = applicables.get(0);
			Method _2nd = applicables.get(1);
			if(cmp.compare(_1st, _2nd)==0){
				throw new RuntimeException(String.format(
								"I cannot choose the proper method to execute : it's ambiguous between at least \n%1$s\n%2$s ", 
								_1st, _2nd));
			}
		}
		return applicables.get(0);
	}
	
	private static List<Method> filterMethods(Class<?> clazz, String name, Filter<Method> filter){
		List<Method> ret = new ArrayList<Method>();
		List<Method> possibilities = getAllMethods(clazz).get(name);

		if(possibilities == null || possibilities.isEmpty()){
			throw new RuntimeException("Method "+name+" not found in "+clazz);
		}
		
		for(Method m : possibilities){
			if(filter.isCompliant(m)){
				ret.add(m);
			}
		}
		
		return ret;
	}
	
	private static List<Class<?>> classesOf(List<?> actualParams) {
		List<Class<?>> actualParamTypes = new ArrayList<Class<?>>();
		for(Object param : actualParams){
			actualParamTypes.add(param==null ? null : param.getClass());
		}
		return actualParamTypes;
	}
	
	public static List<Object> matchArgs(List<? extends MetaMatcher> args){
		List<Object> matchedArgs = new ArrayList<Object>();
		for(MetaMatcher matcher : args){
			matchedArgs.add(matcher.match());
		}
		return matchedArgs;
	}
	/**
	 * Returns true if the given list of expected param types matches the actual parameters
	 * @param paramTypes the params expected by the interface
	 * @param actualParams the params wanted by the caller
	 * @return true if the number of arguments is the same AND all params in order can be assigned to a value of the corresponding type in paramTypes
	 */
	public static boolean argumentsMatch(Class<?>[] paramTypes, List<?> actualParams){
		List<Class<?>> actualParamTypes = classesOf(actualParams);
		return argumentsTypeMatch(paramTypes, actualParamTypes);
	}
	public static boolean argumentsTypeMatch(Class<?>[] expectedParamTypes, List<Class<?>> actualParamTypes){
		if(actualParamTypes.size()!=expectedParamTypes.length) return false;
		
		boolean isMatch = true;
		for(int i=0; i<expectedParamTypes.length; i++){
			Class<?> arg = actualParamTypes.get(i);
			if(arg==null) continue;
			isMatch &= isAssignableFrom(expectedParamTypes[i],arg);
		}
		return isMatch;
	}
	
	/**
	 * Determines if a type a can be assigned/auto-boxed to a type b
	 * @param a a type
	 * @param b a type
	 * @return true if a can be assigned/auto-boxed to type b, false otherwise
	 */
	public static boolean isAssignableFrom(Class<?> a, Class<?> b){
		boolean assignable = false;
		if(a.isPrimitive() && b.isPrimitive()){
			assignable = BOXING_MATRIX.get(a).contains(b);
		} else if(Number.class.equals(a) || Number.class.equals(b)){
			assignable = a.isAssignableFrom(wrap(b));
		} else if((a.isPrimitive() ^ b.isPrimitive()) || (Number.class.isAssignableFrom(a) && Number.class.isAssignableFrom(b))){
			assignable = isAssignableFrom(unwrap(a), unwrap(b));
		} else {
			assignable = a.isAssignableFrom(b);
		}
		return assignable;
	}
	/**
	 * Returns the native type corresponding to type a if a is a primitive type
	 * @param a the type to unwrap
	 * @return true if a is the wrapper to a primitive type
	 */
	public static Class<?> unwrap(Class<?> a){
		if(Character.class.equals(a)){
			return char.class;
		}
		if(Byte.class.equals(a)){
			return byte.class;
		}
		if(Short.class.equals(a)){
			return short.class;
		}
		if(Integer.class.equals(a)){
			return int.class;
		}
		if(Long.class.equals(a)){
			return long.class;
		}
		if(Float.class.equals(a)){
			return float.class;
		}
		if(Double.class.equals(a)){
			return double.class;
		}
		if(Boolean.class.equals(a)){
			return boolean.class;
		}
		return a;
	}
	/**
	 * Returns the wrapper type corresponding to type a if a is a primitive type
	 * @param a the type to wrap
	 * @return the wrapper type corresponding to type a if a is a primitive type
	 */
	public static Class<?> wrap(Class<?> a){
		if(char.class.equals(a)){
			return Character.class;
		}
		if(byte.class.equals(a)){
			return Byte.class;
		}
		if(short.class.equals(a)){
			return Short.class;
		}
		if(int.class.equals(a)){
			return Integer.class;
		}
		if(long.class.equals(a)){
			return Long.class;
		}
		if(float.class.equals(a)){
			return Float.class;
		}
		if(double.class.equals(a)){
			return Double.class;
		}
		if(boolean.class.equals(a)){
			return Boolean.class;
		}
		return a;
	}
	/**
	 * Returns the value o boxed in the type targetType if an autoboxing is possible
	 * @param o the object to autobox
	 * @param targetType the type to which o should be boxed
	 * @return the value o boxed in the type targetType if an autoboxing is possible
	 */
	public static Object autoBox(Object o, Class<?> targetType){
		// not boxable
		if(!(Number.class.isAssignableFrom(targetType) || targetType.isPrimitive())){
			return o;
		}
		if(!(o instanceof Number)){
			return o;
		}
		Number n = (Number)o;
		if(byte.class.equals(unwrap(targetType))){
			return n.byteValue();
		}
		if(short.class.equals(unwrap(targetType))){
			return n.shortValue();
		}
		if(int.class.equals(unwrap(targetType))){
			return n.intValue();
		}
		if(long.class.equals(unwrap(targetType))){
			return n.longValue();
		}
		if(float.class.equals(unwrap(targetType))){
			return n.floatValue();
		}
		if(double.class.equals(unwrap(targetType))){
			return n.doubleValue();
		}
		// how did i end up here ??
		return n;
	}
	
	/**
	 * Applies an autoboxing to all the arguments of a function call
	 * @param args the array of parameters
	 * @param paramTypes the type of the fn params
	 * @return the parameters that all have been autoboxed
	 */
	private static Object[] boxArgs(Object[]args, Class<?>[] paramTypes){
		Object[] ret = new Object[args.length];
		for(int i = 0; i<args.length; i++){
			ret[i] = autoBox(args[i], paramTypes[i]);
		}
		return ret;
	}
	
	/**
	 * Applies an autoboxing to all the arguments of a function call
	 * @param matchers the array of parameters
	 * @param paramTypes the type of the fn params
	 * @return the parameters that all have been autoboxed
	 */
	public static List<MetaMatcher> boxMatchers(List<? extends MetaMatcher> matchers, Class<?>[] paramTypes){
		List<MetaMatcher> ret = new ArrayList<MetaMatcher>();
		for(int i = 0; i<matchers.size(); i++){
			ret.add(boxMatcher(matchers.get(i), paramTypes[i]));
		}
		return ret;
	}
	private static MetaMatcher boxMatcher(MetaMatcher metaMatcher, Class<?> targetType) {
		metaMatcher.setDeclaredType(targetType);
		metaMatcher.setActualValue(autoBox(metaMatcher.getActualValue(), targetType));
		return metaMatcher;
	}
}
